/*************************************************************************
|
|  DESCRIPTION :
|  An example how to program the AT45 Data Flash on the iMX51 EVB.
|  Maybe used as framework for programming other serial and NAND flash.
|
|*************************************************************************/

/*************************************************************************
|  INCLUDES
|*************************************************************************/

#include <string.h>

/*************************************************************************
|  DEFINES
|*************************************************************************/

#define FALSE               0
#define TRUE                1

#define SPI_RXDATA      (*(volatile unsigned long *) (0x70010000))
#define SPI_TXDATA      (*(volatile unsigned long *) (0x70010004))
#define SPI_CONTROL     (*(volatile unsigned long *) (0x70010008))
#define SPI_CONFIG      (*(volatile unsigned long *) (0x7001000C))
#define SPI_INT         (*(volatile unsigned long *) (0x70010010))
#define SPI_DMA         (*(volatile unsigned long *) (0x70010014))
#define SPI_STATUS      (*(volatile unsigned long *) (0x70010018))
#define SPI_PERIOD      (*(volatile unsigned long *) (0x7001001C))
#define SPI_TEST        (*(volatile unsigned long *) (0x70010020))

#define SPI_CTRL_SS1  0x00047021  // control value for SS1 */
#define SPI_CTRL_XCH  (1L<<2)

#define SPI_STAT_TC   (1L<<7)
#define SPI_STAT_RO   (1L<<6)
#define SPI_STAT_RF   (1L<<5)
#define SPI_STAT_RDR  (1L<<4)
#define SPI_STAT_RR   (1L<<3)
#define SPI_STAT_TF   (1L<<2)
#define SPI_STAT_TDR  (1L<<1)
#define SPI_STAT_TE   (1L<<0)

#define AT45_PAGE_SIZE    528
#define AT45_STATUS_RDY   (1<<7)


/*************************************************************************
|  TYPEDEFS
|*************************************************************************/

/* Definition der Standard Typen */
typedef signed char     SINT8;
typedef signed short    SINT16;
typedef signed long     SINT32;
typedef unsigned char   UINT8;
typedef unsigned short  UINT16;
typedef unsigned long   UINT32;
typedef int             BOOL;

typedef unsigned char   CHAR;
typedef unsigned char   BYTE;
typedef unsigned short  WORD;
typedef unsigned long   DWORD;


/*************************************************************************
| Helper for AT45 programming on iMX51 (started from the host)
|*************************************************************************/

/*
Use the following to disassemble:
arm-eabi-objdump --disassemble helper.x > helper.txt

!!!!! use *.x, not *.o because of relative jumpes !!!!!!
*/

typedef struct {
           BYTE*   data;
           UINT32  at45Cmd;
           UINT32  spiCtrl;
  volatile UINT32* pTxFifo;
  volatile UINT32* pControl;
  volatile UINT32* pStatus;
} ParamAT45T;


void AT45_BufferWrite(ParamAT45T* param)
{
  int       part;
  int       i;
  BYTE*     pData;
  DWORD     flashData;

  // write data to buffer 1
  pData = param->data;
  for (part = 0; part < 4; part++) {
    *param->pControl = param->spiCtrl;
    *param->pTxFifo  = param->at45Cmd;
    for (i = 0; i < 33; i++) {
      flashData  = (DWORD)(*pData++) << 24;
      flashData += (DWORD)(*pData++) << 16;
      flashData += (DWORD)(*pData++) << 8;
      flashData += (DWORD)(*pData++);
      *param->pTxFifo = flashData;
    } /* for */
    *param->pStatus  = (SPI_STAT_TC | SPI_STAT_RO);
    *param->pControl = param->spiCtrl | SPI_CTRL_XCH;
    param->at45Cmd  += 132;

    // wait for SPI transfer completed
    while ((*param->pStatus & SPI_STAT_TC) == 0);
  }
  __asm("bkpt");
}


/****************************************************************************
 ****************************************************************************

    AT45_ReadStatus :

 ****************************************************************************/

static BYTE AT45_ReadStatus(void)
{
  DWORD     spiControl;
  DWORD     spiStatus;
  DWORD     flashStatus;

  /* start a 2 byte SPI transfer sequence */
  spiControl = SPI_CTRL_SS1 | ((16-1) << 20);
  SPI_CONTROL = spiControl;
  SPI_TXDATA  = 0xD700;
  SPI_STATUS  = (SPI_STAT_TC | SPI_STAT_RO);
  SPI_CONTROL = spiControl | SPI_CTRL_XCH;

  /* wait for SPI transfer completed */
  do {
    spiStatus = SPI_STATUS;
  } while ((spiStatus & SPI_STAT_TC) == 0);

  /* read data from RX Fifo and display */
  flashStatus = SPI_RXDATA;
  return (BYTE)flashStatus;
} /* AT45_ReadStatus */


/****************************************************************************
 ****************************************************************************

    AT45_PageProg :
    Program one page via Buffer Write and then Buffer to Memory Page

 ****************************************************************************/

static void AT45_PageProg(BOOL erase, DWORD page, const BYTE* data)
{
  int       part;
  int       i;
  DWORD     spiControl;
  DWORD     spiStatus;
  DWORD     spiCmdAddr;
  DWORD     flashData;
  BYTE      flashStatus;

  /*  write data to buffer 1 */
  spiCmdAddr = 0x84000000;
  spiControl = SPI_CTRL_SS1 | (((136*8)-1) << 20);
  for (part = 0; part < 4; part++) {
    SPI_CONTROL = spiControl;
    SPI_TXDATA  = spiCmdAddr;
    for (i = 0; i < 33; i++) {
      flashData  = (DWORD)(*data++) << 24;
      flashData += (DWORD)(*data++) << 16;
      flashData += (DWORD)(*data++) << 8;
      flashData += (DWORD)(*data++);
      SPI_TXDATA = flashData;
    } /* for */
    SPI_STATUS  = (SPI_STAT_TC | SPI_STAT_RO);
    SPI_CONTROL = spiControl | SPI_CTRL_XCH;
    spiCmdAddr += 132;

    /* wait for SPI transfer completed */
    do {
      spiStatus = SPI_STATUS;
    } while ((spiStatus & SPI_STAT_TC) == 0);

  } /* for */

  /* program from buffer to main memory */
  spiCmdAddr  = (erase) ? 0x83000000 : 0x88000000;
  spiCmdAddr |= (page << 10);
  spiControl  = SPI_CTRL_SS1 | (((4*8)-1) << 20);
  SPI_CONTROL = spiControl;
  SPI_TXDATA  = spiCmdAddr;
  SPI_STATUS  = (SPI_STAT_TC | SPI_STAT_RO);
  SPI_CONTROL = spiControl | SPI_CTRL_XCH;

  /* wait for SPI transfer completed */
  do {
      spiStatus = SPI_STATUS;
  } while ((spiStatus & SPI_STAT_TC) == 0);

  /* reset SPI in order to drop received data */
  SPI_CONTROL = SPI_CTRL_SS1 & ~1L;

  /* wait until flash ready */
  do {
    flashStatus = AT45_ReadStatus();
  } while ((flashStatus & AT45_STATUS_RDY) == 0);

} /* AT45_PageProg */


/****************************************************************************
 ****************************************************************************

    AT45_ProgramBlock :
    Program a block of data into the AT45 Data Flash

 ****************************************************************************/

static void AT45_ProgramBlock(BOOL erase, DWORD page, DWORD count, const BYTE* data)
{
  BYTE    pageData[AT45_PAGE_SIZE];
  DWORD   copyCount;

  /* fill spare bytes */
  pageData[512] = 0x00;
  pageData[513] = 0x11;
  pageData[514] = 0x22;
  pageData[515] = 0x33;
  pageData[516] = 0x44;
  pageData[517] = 0x55;
  pageData[518] = 0x66;
  pageData[519] = 0x77;
  pageData[520] = 0x88;
  pageData[521] = 0x99;
  pageData[522] = 0xaa;
  pageData[523] = 0xbb;
  pageData[524] = 0xcc;
  pageData[525] = 0xdd;
  pageData[526] = 0xee;
  pageData[527] = 0xff;

  /* program file */
  do {
    memset(pageData, 0xff, 512);
    copyCount = (count > 512) ? 512 : count;
    memcpy(pageData, data, copyCount);
    AT45_PageProg(erase, page, pageData);
    page++;
    count -= copyCount;
    data  += copyCount;
  } while (count > 0) ;

} /* AT45_ProgramBlock */


/*************************************************************************

 main:

 With the following Telnet sequence the AT45 gets programmed

 load 0xa0100000 e:/temp/dump256k.bin bin
 load 0 m:/temp/helper.x elf
 mm 0xa0007000 0xa0100000
 mm 0xa0007004 127185
 mm 0xa0007008 1024
 go
 ... wait until target enters debug mode

|*************************************************************************/

typedef struct {
        BYTE*   buffer;
        UINT32  count;
        UINT32  page;
} RunParamT;

static BYTE*      flashData = (BYTE*)(0xa0100000);
static RunParamT* pParam    = (RunParamT*)(0xa0007000);

int main(void)
{
  int i;

  // AT45_ProgramBlock(TRUE, 2048, 0x100000, flashData);
  AT45_ProgramBlock(TRUE, pParam->page, pParam->count, pParam->buffer);

  __asm("bkpt");
  return 0;
}
